/* wav.c
   Routines for handling .wav files

   Part of the caiviar package.
   
   Copyright (c) 2001 Matthias Kramm <kramm@quiss.org>

   This file is distributed under the GPL, see file COPYING for details */

#include "../config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "wav.h"

struct WAVBlock {
    char id[4];
    unsigned int size;
};

struct reader {
    FILE*fi;
    unsigned char*data;
    int len;
    int pos;
};

int reader_read(void*mem, int size, int num, struct reader*read) {
    if(read->fi)
	return fread(mem,size,num,read->fi);
    else {
	int n = size*num;
	if(read->pos+n > read->len) {
	    n = read->len - read->pos;
	    n/=size; n*=size;
	}
	memcpy(mem, &read->data[read->pos],n);
	read->pos += n;
	return n/size;
    }
}
void reader_goto(struct reader*read, int pos)
{
    if(read->fi) {
	fseek(read->fi, pos, SEEK_SET);
    } else {
	read->pos = pos;
    }
}
int reader_getlen(struct reader*read)
{
    if(read->fi) {
	int filesize;
	fseek(read->fi, 0, SEEK_END);
	filesize = ftell(read->fi);
	fseek(read->fi, 0, SEEK_SET);
	return filesize;
    } else {
	return read->len;
    }
}
int reader_getpos(struct reader*read)
{
    if(read->fi) {
	return ftell(read->fi);
    } else {
	return read->pos;
    }
}

int getWAVBlock(struct reader*fi, struct WAVBlock*block)
{
    unsigned int size;
    unsigned char b[4];
    if(reader_read(block->id,1,4,fi)<4)
	return 0;
    if(reader_read(b,1,4,fi)<4)
	return 0;
    block->size = b[0]|b[1]<<8|b[2]<<16|b[3]<<24;
    /*printf("Chunk: [%c%c%c%c] (%d bytes)\n", 
	    block->id[0],block->id[1],
	    block->id[2],block->id[3],
	    block->size);*/
    return 1;
}


int readWAV2(struct reader*fi, struct WAV*wav)
{
    unsigned char b[16];
    long int filesize;
    struct WAVBlock block;
    long int pos = 0;
    filesize = reader_getlen(fi);
    memset(wav,0,sizeof(struct WAV));

    //printf("Filesize: %d\n", filesize);

    if(!getWAVBlock (fi, &block)) {
	fprintf(stderr, "readWAV: Couldn't get first block\n");
	return 0;
    }
    if(strncmp(block.id,"RIFF",4)) {
	fprintf(stderr, "readWAV: not a wav file\n");
	return 0;
    }
    if(block.size + 8 < filesize)
	fprintf(stderr, "readWAV: warning - more tags (%d extra bytes)\n",
		filesize - block.size - 8);
    if(block.size + 8 > filesize)
	fprintf(stderr, "readWAV: warning - short file (%d bytes missing)\n",
		block.size + 8 -  filesize);
    if(reader_read(b, 1, 4, fi) < 4) {
	return 0;
    }
    if(strncmp((char*)b, "WAVE", 4)) {
	fprintf(stderr, "readWAV: not a WAV file (2)\n");
	return 0;
    }
 
    do
    {
	if(!getWAVBlock(fi, &block)) {
	    break;
	}

	pos = reader_getpos(fi);
	if(!strncmp(block.id, "fmt ", 4)) {
	    if(reader_read(&b, 1, 16, fi)<16) {
		fprintf(stderr, "readWAV: Couldn't get fmt block\n");
		return 0;
	    }
	    wav->tag = b[0]|b[1]<<8;
	    wav->channels = b[2]|b[3]<<8;
	    wav->sampsPerSec = b[4]|b[5]<<8|b[6]<<16|b[7]<<24;
	    wav->bytesPerSec = b[8]|b[9]<<8|b[10]<<16|b[11]<<24;
	    wav->align = b[12]|b[13]<<8;
	    wav->bps = b[14]|b[15]<<8;
	} else if (!strncmp(block.id, "LIST", 4)) {
	    // subchunk ICMT (comment) may exist
	} else if (!strncmp(block.id, "data", 4)) {
	    wav->data = (unsigned char*)malloc(block.size);
	    if(!wav->data) {
		fprintf(stderr, "Out of memory (%d bytes needed)", block.size);
		return 0;
	    }
	    if(reader_read(wav->data, 1, block.size, fi) < block.size) {
		fprintf(stderr, "readWAV: Couldn't get data block\n");
		return 0;
	    }
	    wav->size = block.size;
	}
	pos+=block.size;
	reader_goto(fi, pos);
    }
    while (pos < filesize);
    if(!wav->data)
	return 0;

    return 1;
}

int writeWAV(char*filename, struct WAV*wav)
{
    FILE*fi = fopen(filename, "wb");
    char*b="RIFFWAVEfmt \x10\0\0\0data";
    char c[16];
    unsigned long int w32;
    if(!fi)
	return 0;
    fwrite(b, 4, 1, fi);
    w32=(/*fmt*/8+0x10+/*data*/8+wav->size);
    c[0] = w32;
    c[1] = w32>>8;
    c[2] = w32>>16;
    c[3] = w32>>24;
    fwrite(c, 4, 1, fi);
    fwrite(&b[4], 12, 1, fi);
    c[0] = wav->tag;
    c[1] = wav->tag>>8;
    c[2] = wav->channels;
    c[3] = wav->channels>>8;
    c[4] = wav->sampsPerSec;
    c[5] = wav->sampsPerSec>>8;
    c[6] = wav->sampsPerSec>>16;
    c[7] = wav->sampsPerSec>>24;
    c[8] = wav->bytesPerSec;
    c[9] = wav->bytesPerSec>>8;
    c[10] = wav->bytesPerSec>>16;
    c[11] = wav->bytesPerSec>>24;
    c[12] = wav->align;
    c[13] = wav->align>>8;
    c[14] = wav->bps;
    c[15] = wav->bps>>8;
    fwrite(c, 16, 1, fi);
    fwrite(&b[16], 4, 1, fi);
    c[0] = wav->size;
    c[1] = wav->size>>8;
    c[2] = wav->size>>16;
    c[3] = wav->size>>24;
    fwrite(c,4,1,fi);
    fwrite(wav->data,wav->size,1,fi);
    fclose(fi);
    return 1;
}

void printWAVInfo(struct WAV*wav)
{
    printf("tag:%04x channels:%d samples/sec:%d bytes/sec:%d align:%d bits/sample:%d size:%d\n",
	    wav->tag, wav->channels, wav->sampsPerSec, wav->bytesPerSec, 
	    wav->align, wav->bps, wav->size);
}

int convertWAV2mono(struct WAV*src, struct WAV*dest, int rate)
{
    int samplelen=src->size/src->align;
    int bps=src->bps;
    double ratio;
    double pos = 0;
    int pos2 = 0;
    int channels=src->channels;
    int i;
    int fill;

    dest->sampsPerSec = rate;
    dest->bps = 16;
    dest->channels = 1;
    dest->align = 2;
    dest->tag = src->tag;
    dest->bytesPerSec = dest->sampsPerSec*dest->align;

    ratio = (double)dest->sampsPerSec/(double)src->sampsPerSec;
    fill = (int)(ratio+1)*2;
    
    dest->data = (unsigned char*)malloc((int)(samplelen*ratio*2)+128);
    if(!dest->data) 
	return 0;
    dest->size = (int)(samplelen*ratio)*2;

    if(bps == 8) {
	if(ratio <= 1) {
	    for(i=0; i<src->size; i+=channels) {
		int pos2 = ((int)pos)*2;
		dest->data[pos2+0] = 0;
		dest->data[pos2+1] = src->data[i]+128;
		pos += ratio;
	    }
	} else {
	    for(i=0; i<src->size; i+=channels) {
		int j;
		int pos2 = ((int)pos)*2;
		for(j=0;j<fill;j+=2) {
		    dest->data[pos2+j+0] = 0;
		    dest->data[pos2+j+1] = src->data[i]+128;
		}
		pos += ratio;
	    }
	}
    } else if(bps == 16) {
	if(ratio <= 1) {
	    for(i=0; i<src->size/2; i+=channels) {
		int pos2 = ((int)pos)*2;
		dest->data[pos2+0]=src->data[i*2];
		dest->data[pos2+1]=src->data[i*2+1];
		pos += ratio;
	    }
	} else {
	    for(i=0; i<src->size/2; i+=channels) {
		int j;
		int pos2 = ((int)pos)*2;
		for(j=0;j<fill;j+=2) {
		    dest->data[pos2+j+0] = src->data[i*2];
		    dest->data[pos2+j+1] = src->data[i*2+1];
		}
		pos += ratio;
	    }
	}
    } else if(bps == 32) {
	if(ratio <= 1) {
	    for(i=0; i<src->size/4; i+=channels) {
		int pos2 = ((int)pos)*2;
		dest->data[pos2+0]=src->data[i*4+2];
		dest->data[pos2+1]=src->data[i*4+3];
		pos += ratio;
	    }
	} else {
	    for(i=0; i<src->size/4; i+=channels) {
		int j;
		int pos2 = ((int)pos)*2;
		for(j=0;j<fill;j+=2) {
		    dest->data[pos2+j+0] = src->data[i*4+2];
		    dest->data[pos2+j+1] = src->data[i*4+3];
		}
		pos += ratio;
	    }
	}
    }
    return 1;
}

unsigned char imap[256] = {
0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8,
0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC,
0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2,
0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA,
0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6,
0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1,
0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9,
0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5,
0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED,
0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB,
0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7,
0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF,
0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};

unsigned char wavsample2law(short int val) 
{
    char sign;
    unsigned char carry;
    unsigned char alaw;

    sign=val>=0?1:0;
    if(val<0)
	val*=-1;

    if (val < 0x0800) {
        if (val < 0x0200) {
            if (val < 0x0100) {
		carry = (val&12)?1:0;
                val&=0xF0; val>>=3;
		alaw=0x00+val+carry;
            } else {
		carry = (val&28)?1:0;
                val&=0xE0; val>>=4;
		alaw=0x20+val+carry;
            }
        } else {
            if (val < 0x0400)
            {
		carry = (val&60)?1:0;
                val&=0x1C0; val>>=5;
		alaw=0x30+val+carry;
            } else {
		carry = (val&124)?1:0;
                val&=0x380; val>>=6;
		alaw=0x40+val+carry;
            }
        }
    } else {
        if (val < 0x2000) {
            if (val < 0x1000) {
		carry = (val&252)?1:0;
                val&=0x700; val>>=7;
		alaw=0x50+val+carry;
            } else {
		carry = (val&508)?1:0;
                val&=0xE00; val>>=8;
		alaw=0x60+val+carry;

            }
        } else {
            if (val < 0x4000)
            {
		carry = (val&1020)?1:0;
                val&=0x1C00; val>>=9;
		alaw=0x70+val+carry;

            } else {
		alaw = 0x7f;
	    }
        }
    }
    alaw += sign*0x80;
    return imap[(alaw ^ 0x55)];
}

short int law2wav(unsigned char value)
{
    unsigned char segment;
    unsigned char offset;
    short int pcm;
    value = imap[value];
    value^=0x55;
    segment = value & 0x70;
    segment >>= 4;
    offset=value & 0x0f;
    if (segment < 0x04) {
        if (segment < 0x02) {
            if (segment < 0x01)
                pcm=(0x0002+(offset<<3));
            else
                pcm=(0x0082+(offset<<3));
        } else {
            if (segment < 0x03)
                pcm=(0x0102+(offset<<4));
            else
                pcm=(0x0202+(offset<<5));
        }
    } else {
        if (segment < 0x06) {
            if (segment < 0x05)
                pcm=(0x0402+(offset<<6));
            else
                pcm=(0x0802+(offset<<7));
        } else {
            if (segment < 0x07)
                pcm=(0x1002+(offset<<8));
            else
                pcm=(0x2002+(offset<<9));
        }
    }
    if (!(value&0x80))
        pcm*=-1;
    return pcm;
}


void convertWAV2LAW(struct WAV*src, unsigned char**data, int*size)
{
    struct WAV mono;
    int t;
    convertWAV2mono(src, &mono, 8000);
    *size = mono.size/2;
    *data = (unsigned char*)malloc(*size);

    for(t=0;t<mono.size/2;t++) {
	signed short value = ((signed short*)mono.data)[t];
	(*data)[t]=wavsample2law(value);
    }
    free(mono.data);
}

void convertLAW2WAV(void*data, int size, struct WAV*dest)
{
    unsigned char*d = (unsigned char*)data;
    int t;

    dest->data = (unsigned char*)malloc(size*2);
    dest->size = size*2;
    dest->sampsPerSec = 8000;
    dest->bps = 16;
    dest->channels = 1;
    dest->align = 2;
    dest->tag = 0x0001;
    dest->bytesPerSec = dest->sampsPerSec*dest->align;

    for(t=0;t<size;t++) {
	((signed short*)dest->data)[t] = law2wav(*d);
	d++;
    }
}

int readWAV(char* filename, struct WAV*wav)
{
    struct reader r;
    FILE*fi = fopen(filename, "rb");
    if(!fi) {
	fprintf(stderr, "Couldn't open %s\n", filename);
	return 0;
    }
    r.fi = fi;
    r.data = 0;
    r.len = 0;
    r.pos = 0;
    return readWAV2(&r,wav);
}

int readWAVFromMem(void*data, int len, struct WAV*wav)
{
    struct reader r;
    r.fi = 0;
    r.data = (unsigned char*)data;
    r.len = len;
    r.pos = 0;
    return readWAV2(&r,wav);
}

unsigned char*wav2law(unsigned char*wavedata, int len, int *len2)
{
    struct WAV wav;
    unsigned char* data;
    int size;
    if(!readWAVFromMem(wavedata, len, &wav))
	return 0;
    convertWAV2LAW(&wav, &data, &size);
    free(wav.data);
    *len2 = size;
    return data;
}

